@@ -12,7 +12,8 @@ use Cwd;
 my $builddir = getcwd;
 
 
-@EXPORT = qw(new_memcached sleep mem_get_is mem_gets mem_gets_is mem_stats free_port);
+@EXPORT = qw(new_memcached sleep mem_get_is mem_gets mem_gets_is mem_stats
+             supports_sasl free_port);
 
 sub sleep {
     my $n = shift;
@@ -141,6 +142,12 @@ sub supports_udp {
     return 1;
 }
 
+sub supports_sasl {
+    my $output = `$builddir/memcached-debug -h`;
+    return 1 if $output =~ /sasl/i;
+    return 0;
+}
+
 sub new_memcached {
     my ($args, $passed_port) = @_;
     my $port = $passed_port || free_port();
@@ -165,6 +172,7 @@ sub new_memcached {
     if ($< == 0) {
         $args .= " -u root";
     }
+
     my $childpid = fork();
 
     my $exe = "$builddir/memcached-debug";
