@@ -3,6 +3,7 @@
 #include <pthread.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/wait.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <netinet/in.h>
@@ -13,15 +14,14 @@
 #include <errno.h>
 #include <assert.h>
 #include <string.h>
-#include <inttypes.h>
-#include <stdbool.h>
 #include <unistd.h>
 #include <netinet/in.h>
+#include <fcntl.h>
 
-#include "protocol_binary.h"
 #include "config.h"
 #include "cache.h"
 #include "util.h"
+#include "protocol_binary.h"
 
 #define TMP_TEMPLATE "/tmp/test_file.XXXXXXX"
 
@@ -114,6 +114,39 @@ static enum test_return cache_reuse_test(void)
     return TEST_PASS;
 }
 
+
+static enum test_return cache_bulkalloc(size_t datasize)
+{
+    cache_t *cache = cache_create("test", datasize, sizeof(char*),
+                                  NULL, NULL);
+#define ITERATIONS 1024
+    void *ptr[ITERATIONS];
+
+    for (int ii = 0; ii < ITERATIONS; ++ii) {
+        ptr[ii] = cache_alloc(cache);
+        assert(ptr[ii] != 0);
+        memset(ptr[ii], 0xff, datasize);
+    }
+
+    for (int ii = 0; ii < ITERATIONS; ++ii) {
+        cache_free(cache, ptr[ii]);
+    }
+
+#undef ITERATIONS
+    cache_destroy(cache);
+    return TEST_PASS;
+}
+
+static enum test_return test_issue_161(void)
+{
+    enum test_return ret = cache_bulkalloc(1);
+    if (ret == TEST_PASS) {
+        ret = cache_bulkalloc(512);
+    }
+
+    return ret;
+}
+
 static enum test_return cache_redzone_test(void)
 {
 #ifndef HAVE_UMEM_H
@@ -157,6 +190,7 @@ static enum test_return test_safe_strtoul(void) {
     assert(val == 123);
     assert(!safe_strtoul("", &val));  // empty
     assert(!safe_strtoul("123BOGUS", &val));  // non-numeric
+    assert(!safe_strtoul(" issue221", &val));  // non-numeric
     /* Not sure what it does, but this works with ICC :/
        assert(!safe_strtoul("92837498237498237498029383", &val)); // out of range
     */
@@ -181,6 +215,7 @@ static enum test_return test_safe_strtoull(void) {
     assert(!safe_strtoull("", &val));  // empty
     assert(!safe_strtoull("123BOGUS", &val));  // non-numeric
     assert(!safe_strtoull("92837498237498237498029383", &val)); // out of range
+    assert(!safe_strtoull(" issue221", &val));  // non-numeric
 
     // extremes:
     assert(safe_strtoull("18446744073709551615", &val)); // 2**64 - 1
@@ -201,6 +236,7 @@ static enum test_return test_safe_strtoll(void) {
     assert(!safe_strtoll("", &val));  // empty
     assert(!safe_strtoll("123BOGUS", &val));  // non-numeric
     assert(!safe_strtoll("92837498237498237498029383", &val)); // out of range
+    assert(!safe_strtoll(" issue221", &val));  // non-numeric
 
     // extremes:
     assert(!safe_strtoll("18446744073709551615", &val)); // 2**64 - 1
@@ -229,6 +265,7 @@ static enum test_return test_safe_strtol(void) {
     assert(!safe_strtol("", &val));  // empty
     assert(!safe_strtol("123BOGUS", &val));  // non-numeric
     assert(!safe_strtol("92837498237498237498029383", &val)); // out of range
+    assert(!safe_strtol(" issue221", &val));  // non-numeric
 
     // extremes:
     /* This actually works on 64-bit ubuntu
@@ -271,7 +308,8 @@ static pid_t start_server(in_port_t *port_out, bool daemon, int timeout) {
        overwrite each other
     */
     char coreadm[128];
-    sprintf(coreadm, "coreadm -p core.%%f.%%p %lu", (unsigned long)getpid());
+    snprintf(coreadm, sizeof(coreadm),
+             "coreadm -p core.%%f.%%p %lu", (unsigned long)getpid());
     system(coreadm);
 #endif
 
@@ -398,7 +436,7 @@ static struct addrinfo *lookuphost(const char *hostname, in_port_t port)
     return ai;
 }
 
-static int connect_server(const char *hostname, in_port_t port)
+static int connect_server(const char *hostname, in_port_t port, bool nonblock)
 {
     struct addrinfo *ai = lookuphost(hostname, port);
     int sock = -1;
@@ -410,6 +448,14 @@ static int connect_server(const char *hostname, in_port_t port)
                      strerror(errno));
              close(sock);
              sock = -1;
+          } else if (nonblock) {
+              int flags = fcntl(sock, F_GETFL, 0);
+              if (flags < 0 || fcntl(sock, F_SETFL, flags | O_NONBLOCK) < 0) {
+                  fprintf(stderr, "Failed to enable nonblocking mode: %s\n",
+                          strerror(errno));
+                  close(sock);
+                  sock = -1;
+              }
           }
        } else {
           fprintf(stderr, "Failed to create socket: %s\n", strerror(errno));
@@ -514,7 +560,7 @@ static enum test_return test_issue_92(void) {
     char buffer[1024];
 
     close(sock);
-    sock = connect_server("127.0.0.1", port);
+    sock = connect_server("127.0.0.1", port, false);
 
     send_ascii_command("stats cachedump 1 0 0\r\n");
     read_ascii_response(buffer, sizeof(buffer));
@@ -525,13 +571,62 @@ static enum test_return test_issue_92(void) {
     assert(strncmp(buffer, "CLIENT_ERROR", strlen("CLIENT_ERROR")) == 0);
 
     close(sock);
-    sock = connect_server("127.0.0.1", port);
+    sock = connect_server("127.0.0.1", port, false);
+    return TEST_PASS;
+}
+
+static enum test_return test_issue_102(void) {
+    char buffer[4096];
+    memset(buffer, ' ', sizeof(buffer));
+    buffer[sizeof(buffer) - 1] = '\0';
+
+    close(sock);
+    sock = connect_server("127.0.0.1", port, false);
+
+    send_ascii_command(buffer);
+    /* verify that the server closed the connection */
+    assert(read(sock, buffer, sizeof(buffer)) == 0);
+    close(sock);
+    sock = connect_server("127.0.0.1", port, false);
+
+    snprintf(buffer, sizeof(buffer), "gets ");
+    size_t offset = 5;
+    while (offset < 4000) {
+        offset += snprintf(buffer + offset, sizeof(buffer) - offset,
+                           "%010u ", (unsigned int)offset);
+    }
+
+    send_ascii_command(buffer);
+    usleep(250);
+
+    send_ascii_command("\r\n");
+    char rsp[80];
+    read_ascii_response(rsp, sizeof(rsp));
+    assert(strncmp(rsp, "END", strlen("END")) == 0);
+    buffer[3]= ' ';
+    send_ascii_command(buffer);
+    usleep(250);
+    send_ascii_command("\r\n");
+    read_ascii_response(rsp, sizeof(rsp));
+    assert(strncmp(rsp, "END", strlen("END")) == 0);
+
+    memset(buffer, ' ', sizeof(buffer));
+    int len = snprintf(buffer + 101, sizeof(buffer) - 101, "gets foo");
+    buffer[101 + len] = ' ';
+    buffer[sizeof(buffer) - 1] = '\0';
+    send_ascii_command(buffer);
+    /* verify that the server closed the connection */
+    assert(read(sock, buffer, sizeof(buffer)) == 0);
+
+    close(sock);
+    sock = connect_server("127.0.0.1", port, false);
+
     return TEST_PASS;
 }
 
 static enum test_return start_memcached_server(void) {
     server_pid = start_server(&port, false, 600);
-    sock = connect_server("127.0.0.1", port);
+    sock = connect_server("127.0.0.1", port, false);
     return TEST_PASS;
 }
 
@@ -728,6 +823,33 @@ static off_t flush_command(char* buf, size_t bufsz, uint8_t cmd, uint32_t exptim
     return size;
 }
 
+
+static off_t touch_command(char* buf,
+                           size_t bufsz,
+                           uint8_t cmd,
+                           const void* key,
+                           size_t keylen,
+                           uint32_t exptime) {
+    protocol_binary_request_touch *request = (void*)buf;
+    assert(bufsz > sizeof(*request));
+
+    memset(request, 0, sizeof(*request));
+    request->message.header.request.magic = PROTOCOL_BINARY_REQ;
+    request->message.header.request.opcode = cmd;
+
+    request->message.header.request.keylen = htons(keylen);
+    request->message.header.request.extlen = 4;
+    request->message.body.expiration = htonl(exptime);
+    request->message.header.request.bodylen = htonl(keylen + 4);
+
+    request->message.header.request.opaque = 0xdeadbeef;
+
+    off_t key_offset = sizeof(protocol_binary_request_no_extras) + 4;
+
+    memcpy(buf + key_offset, key, keylen);
+    return sizeof(protocol_binary_request_no_extras) + 4 + keylen;
+}
+
 static off_t arithmetic_command(char* buf,
                                 size_t bufsz,
                                 uint8_t cmd,
@@ -845,7 +967,8 @@ static void validate_response_header(protocol_binary_response_no_extras *respons
     } else {
         assert(response->message.header.response.cas == 0);
         assert(response->message.header.response.extlen == 0);
-        if (cmd != PROTOCOL_BINARY_CMD_GETK) {
+        if (cmd != PROTOCOL_BINARY_CMD_GETK &&
+            cmd != PROTOCOL_BINARY_CMD_GATK) {
             assert(response->message.header.response.keylen == 0);
         }
     }
@@ -889,7 +1012,7 @@ static enum test_return test_binary_quit_impl(uint8_t cmd) {
     /* Socket should be closed now, read should return 0 */
     assert(read(sock, buffer.bytes, sizeof(buffer.bytes)) == 0);
     close(sock);
-    sock = connect_server("127.0.0.1", port);
+    sock = connect_server("127.0.0.1", port, false);
 
     return TEST_PASS;
 }
@@ -1475,7 +1598,7 @@ static enum test_return test_binary_stat(void) {
 }
 
 static enum test_return test_binary_illegal(void) {
-    uint8_t cmd = 0x1b;
+    uint8_t cmd = 0x25;
     while (cmd != 0x00) {
         union {
             protocol_binary_request_no_extras request;
@@ -1579,12 +1702,25 @@ static enum test_return test_binary_pipeline_hickup_chunk(void *buffer, size_t b
                              key, keylen, NULL, 0);
             break;
 
+        case PROTOCOL_BINARY_CMD_TOUCH:
+        case PROTOCOL_BINARY_CMD_GAT:
+        case PROTOCOL_BINARY_CMD_GATQ:
+        case PROTOCOL_BINARY_CMD_GATK:
+        case PROTOCOL_BINARY_CMD_GATKQ:
+            len = touch_command(command.bytes, sizeof(command.bytes), cmd,
+                                key, keylen, 10);
+            break;
+
         case PROTOCOL_BINARY_CMD_STAT:
             len = raw_command(command.bytes, sizeof(command.bytes),
                               PROTOCOL_BINARY_CMD_STAT,
                               NULL, 0, NULL, 0);
             break;
 
+        case PROTOCOL_BINARY_CMD_SASL_LIST_MECHS:
+        case PROTOCOL_BINARY_CMD_SASL_AUTH:
+        case PROTOCOL_BINARY_CMD_SASL_STEP:
+            /* Ignoring SASL */
         case PROTOCOL_BINARY_CMD_QUITQ:
         case PROTOCOL_BINARY_CMD_QUIT:
             /* I don't want to pass on the quit commands ;-) */
@@ -1641,6 +1777,76 @@ static enum test_return test_binary_pipeline_hickup(void)
     return TEST_PASS;
 }
 
+
+static enum test_return test_issue_101(void) {
+    const int max = 2;
+    enum test_return ret = TEST_PASS;
+    int fds[max];
+    int ii = 0;
+    pid_t child = 0;
+
+    if (getenv("SKIP_TEST_101") != NULL) {
+        return TEST_SKIP;
+    }
+
+    const char *command = "stats\r\nstats\r\nstats\r\nstats\r\nstats\r\n";
+    size_t cmdlen = strlen(command);
+
+    server_pid = start_server(&port, false, 1000);
+
+    for (ii = 0; ii < max; ++ii) {
+        fds[ii] = connect_server("127.0.0.1", port, true);
+        assert(fds[ii] > 0);
+    }
+
+    /* Send command on the connection until it blocks */
+    for (ii = 0; ii < max; ++ii) {
+        bool more = true;
+        do {
+            ssize_t err = write(fds[ii], command, cmdlen);
+            if (err == -1) {
+                switch (errno) {
+                case EINTR:
+                    break;
+                case ENOMEM:
+                case EWOULDBLOCK:
+                    more = false;
+                    break;
+                default:
+                    ret = TEST_FAIL;
+                    goto cleanup;
+                }
+            }
+        } while (more);
+    }
+
+    child = fork();
+    if (child == (pid_t)-1) {
+        abort();
+    } else if (child > 0) {
+        int stat;
+        pid_t c;
+        while ((c = waitpid(child, &stat, 0)) == (pid_t)-1 && errno == EINTR);
+        assert(c == child);
+        assert(stat == 0);
+    } else {
+        sock = connect_server("127.0.0.1", port, false);
+        ret = test_binary_noop();
+        close(sock);
+        exit(0);
+    }
+
+ cleanup:
+    /* close all connections */
+    for (ii = 0; ii < max; ++ii) {
+        close(fds[ii]);
+    }
+
+    assert(kill(server_pid, SIGTERM) == 0);
+
+    return ret;
+}
+
 typedef enum test_return (*TEST_FUNC)(void);
 struct testcase {
     const char *description;
@@ -1654,15 +1860,18 @@ struct testcase testcases[] = {
     { "cache_destructor", cache_destructor_test },
     { "cache_reuse", cache_reuse_test },
     { "cache_redzone", cache_redzone_test },
+    { "issue_161", test_issue_161 },
     { "strtol", test_safe_strtol },
     { "strtoll", test_safe_strtoll },
     { "strtoul", test_safe_strtoul },
     { "strtoull", test_safe_strtoull },
     { "issue_44", test_issue_44 },
     { "vperror", test_vperror },
+    { "issue_101", test_issue_101 },
     /* The following tests all run towards the same server */
     { "start_server", start_memcached_server },
     { "issue_92", test_issue_92 },
+    { "issue_102", test_issue_102 },
     { "binary_noop", test_binary_noop },
     { "binary_quit", test_binary_quit },
     { "binary_quitq", test_binary_quitq },
